"""
This module contains support functions to turn
:class:`~kingdon.multivector.MultiVector`'s into matrices.

This follows the approach outlined in
Graded Symmetry Groups: Plane and Simple, section 10.
See the paper for more details.
"""
from functools import reduce
from itertools import combinations
from typing import Callable
import numpy as np
import sympy


I2 = np.array([[1,0], [0,1]])
Ip2 = np.array([[1,0], [0,-1]])
P2 = np.array([[0,1], [1,0]])
N2 = np.array([[0,1], [-1,0]])
Z2 = np.array([[0,0], [1,0]])


def ordering_matrix(Rs):
    """
    Matrix reps are determined up to similarity transform.
    But not all similarity transforms are equal.
    This function creates the one similarity transform that gives all
    entries in the first column of the matrix a positive sign.
    In doing so, matrix-matrix multiplication is identical to matrix-vector
    multiplication.

    :param Rs: sequence of matrix reps for all the basis blades of an algebra.
    :return: The similarity transform to beat all similarity transforms.
    """
    columns = [Ri[:, 0] for Ri in Rs]
    return np.vstack(columns)


def matrix_rep(p=0, q=0, r=0):
    """
    Create the matrix reps of all the basis blades of an algebra.
    These are selected such that the entries in the first column
    of the matrix have positive sign, and thus matrix-matrix multiplication
    is identical to matrix-vector multiplication.

    :param p: number of positive dimensions.
    :param q: number of negative dimensions.
    :param r: number of null dimensions.
    :return: sequence of matrix reps for the basis-blades.
    """
    d = p + q + r
    I = I2
    P = P2
    Z = Z2
    N = N2
    Ip = Ip2

    # Store all the signature matrices needed for the E_i
    Ss = [Z for _ in range(r)]
    Ss.extend([P for _ in range(p)])
    Ss.extend([N for _ in range(q)])
    # Construct the matrix reps for the E_i from E_d to E_0
    Es = []
    for i, Si in enumerate(Ss):
        mats = [I for _ in range(i)]
        mats.append(Si)
        mats.extend([Ip for _ in range(d - i - 1)])
        Es.append(reduce(np.kron, mats, 1))
    Es = list(reversed(Es))

    Rs = Es.copy()
    Iden = reduce(np.kron, [I for _ in range(d)])
    Rs.insert(0, Iden)

    # Extend Rs with the higher order basis-blades.
    for i in range(2, d+1):
        Rs_grade_i = [reduce(lambda x, y: x @ y, comb)
                      for comb in combinations(Es, r=i)]
        Rs.extend(Rs_grade_i)

    O = ordering_matrix(Rs)
    return [O @ Ri @ O.T for Ri in Rs]


def expr_as_matrix(expr: Callable, *inputs, res_like: "MultiVector" = None):
    """
    This represents any GA expression as a matrix. To illustrate by example, we might want to
    represent the multivector equation y = R >> x as a matrix equation y = Ax.
    To obtain A, call this function as follows::

        alg = Algebra(3, 0, 1)
        R = alg.evenmv(name='R')
        x = alg.vector(name='x')
        A, y = expr_as_matrix(lambda R, x: R >> x, R, x)

    In order to build the matrix rep the input `expr` is evaluated, so make sure the inputs
    to the expression are given in the correct order.
    The last of the positional arguments is assumed to be the vector x.

    :expr: Callable representing a valid GA expression.
        Can also be a :class:`~kingdon.operator_dict.OperatorDict`.
    :inputs: All positional arguments are consider symbolic input arguments to `expr`. The last of these is assumed to
        represent the vector `x` in `y = Ax`.
    :res_like: optional multivector corresponding to the desired output. If None, then the full output is returned.
        However, if only a subsegment of the output is desired, provide a multivector with the desired shape.
        In the example above setting, `res_like = alg.vector(e1=1)` would mean only the e1 component of the matrix
        is returned. This does not have to be a symbolic multivector, only the keys are checked.
    :return: This function returns the matrix representation, and the result of applying the expression to the input.
    """
    x = inputs[-1]
    y = expr(*inputs)
    alg = x.algebra
    if res_like is not None:
        y = alg.multivector({k: sympy.sympify(getattr(y, alg.bin2canon[k])) for k in res_like.keys()})

    A = sympy.zeros(len(y), len(x))
    for i, (blade_y, yi) in enumerate(y.items()):
        cv = sympy.collect(yi.expand(), x.values())
        for j, (blade_x, xj) in enumerate(x.items()):
            A[i, j] = cv.coeff(xj)
    return A, y



import pytest
from kingdon.polynomial import Polynomial, RationalPolynomial, compare


@pytest.mark.parametrize("a, b, res", [
    # Equal polynomials
    (Polynomial([[1, 'x'], [2, 'y'], [3, 'z']]), Polynomial([[1, 'x'], [2, 'y'], [3, 'z']]), 0),
    # a < b
    (Polynomial([[1, 'x'], [2, 'y'], [3, 'z']]), Polynomial([[1, 'x'], [2, 'y'], [4, 'z']]), -1),
    # a > b
    (Polynomial([[1, 'x'], [2, 'y'], [4, 'z']]), Polynomial([[1, 'x'], [2, 'y'], [3, 'z']]), 1),
    # b undefined
    (Polynomial([[1, 'x'], [2, 'y'], [4, 'z']]), None, -1),
    # a undefined
    (None, Polynomial([[1, 'x'], [2, 'y'], [4, 'z']]), 1),
    # Both undefined
    (None, None, 1),
])
def test_compare(a, b, res):
    assert compare(a, b) == res


@pytest.mark.parametrize("a, b, res", [
    # Add single variable polynomials
    (Polynomial([[1, 'x']]), Polynomial([[2, 'x']]), Polynomial([[3, 'x']])),
    # Add multi-variable polynomials
    (Polynomial([[1, 'x', 'y'], [2, 'z']]), Polynomial([[3, 'x', 'y'], [4, 'z']]), Polynomial([[4, 'x', 'y'], [6, 'z']])),
    # Add maintaining sort order
    (Polynomial([[1, 'x'], [2, 'y'], [3, 'z']]), Polynomial([[4, 'x'], [5, 'y'], [6, 'z']]), Polynomial([[5, 'x'], [7, 'y'], [9, 'z']])),
    # Add with zero polynomials
    (Polynomial([[1, 'x'], [2, 'y'], [3, 'z']]), 0, Polynomial([[1, 'x'], [2, 'y'], [3, 'z']])),
    # Add with negative coefficients
    (Polynomial([[1, 'x'], [-2, 'y'], [3, 'z']]), Polynomial([[-1, 'x'], [2, 'y'], [-3, 'z']]), 0),
    # Same denominator
    (RationalPolynomial([[1, 'x'], [2, 'y']], [[1, 'z']]), RationalPolynomial([[3, 'x'], [4, 'y']], [[1, 'z']]),
     RationalPolynomial([[4, 'x'], [6, 'y']], [[1, 'z']])),
    # Different denoms
    (RationalPolynomial([[1, 'x'], [2, 'y']], [[1, 'z']]), RationalPolynomial([[3, 'x'], [4, 'y']], [[2, 'z']]),
     RationalPolynomial([[5, 'x', 'z'], [8, 'y', 'z']], [[2, 'z', 'z']])),
    # Add with zero
    (RationalPolynomial([[1, 'x'], [2, 'y']], [[1, 'z']]), 0,
     RationalPolynomial([[1, 'x'], [2, 'y']], [[1, 'z']])),
    # With negative coefficients
    (RationalPolynomial([[-1, 'x'], [2, 'y']], [[1, 'z']]), RationalPolynomial([[1, 'x'], [-2, 'y']], [[1, 'z']]), 0),
])
def test_add(a, b, res):
    assert a + b == res


@pytest.mark.parametrize("a, b, res", [
    # single variable polynomials
    (Polynomial([[1, 'x']]), Polynomial([[2, 'x']]), Polynomial([[2, 'x', 'x']])),
    # multi-variable polynomials
    (Polynomial([[1, 'x', 'y'], [2, 'z']]), Polynomial([[3, 'x', 'y'], [4, 'z']]),
     Polynomial([[3, 'x', 'x', 'y', 'y'], [10, 'x', 'y', 'z'], [8, 'z', 'z']])),
    # with zero polynomials
    (Polynomial([[1, 'x'], [2, 'y'], [3, 'z']]), 0, 0),
    # With negative coefficients
    (Polynomial([[1, 'x'], [-2, 'y'], [3, 'z']]), Polynomial([[-1, 'x'], [2, 'y'], [-3, 'z']]),
     Polynomial([[-1, 'x', 'x'], [4, 'x', 'y'], [-6, 'x', 'z'],[-4, 'y', 'y'], [12, 'y', 'z'], [-9, 'z', 'z']])),
    (RationalPolynomial([[1, 'x']]), RationalPolynomial([[2, 'x']]), RationalPolynomial([[2, 'x', 'x']])),
    (RationalPolynomial([[1, 'x', 'y'], [2, 'z']], [[1]]), RationalPolynomial([[3, 'x', 'y'], [4, 'z']], [[1]]),
     RationalPolynomial([[3, 'x', 'x', 'y', 'y'], [10, 'x', 'y', 'z'], [8, 'z', 'z']], [[1]])),
    (RationalPolynomial([[1, 'x'], [2, 'y'], [3, 'z']]), 0, 0),
    (RationalPolynomial([[1, 'x'], [2, 'y'], [3, 'z']]), 1, RationalPolynomial([[1, 'x'], [2, 'y'], [3, 'z']])),
    (RationalPolynomial([[2, 'x', 'z']], [[1]]), RationalPolynomial([[1]], [[3, 'z']]),
     RationalPolynomial([[2, 'x',]], [[3]])),
])
def test_mul(a, b, res):
    assert a * b == res


@pytest.mark.parametrize("a, res", [
    # single variable polynomials
    (Polynomial([[1, 'x']]), Polynomial([[-1, 'x']])),
    (Polynomial([[1, 'x', 'y'], [2, 'z']]), Polynomial([[-1, 'x', 'y'], [-2, 'z']])),
    (Polynomial([]), Polynomial([])),
    (Polynomial([[-1, 'x'], [2, 'y'], [-3, 'z']]), Polynomial([[1, 'x'], [-2, 'y'], [3, 'z']])),
])
def test_neg(a, res):
    assert -a == res


@pytest.mark.parametrize("a, res", [
    # single variable polynomials
    (RationalPolynomial([[1, 'x']]), RationalPolynomial([[1]], [[1, 'x']])),
    (RationalPolynomial([[1, 'x', 'y'], [2, 'z']], [[1]]), RationalPolynomial([[1]], [[1, 'x', 'y'], [2, 'z']])),
])
def test_inverse(a, res):
    assert a.inv() == res


@pytest.mark.parametrize("a, res", [
    (RationalPolynomial([[1, 'x']]), RationalPolynomial([[1]])),
    (RationalPolynomial([[1, 'x','x'],[2, 'x', 'y'], [3, 'y', 'y', 'y']]), RationalPolynomial([[1]])),
])
def test_inverse_mul(a, res):
    assert a * a.inv() == res

@pytest.mark.parametrize("a, n, res", [
    (Polynomial([[2, 'x']]), 2, Polynomial([[4, 'x', 'x']])),
    (RationalPolynomial([[2, 'x']]), 2, RationalPolynomial([[4, 'x', 'x']])),
    (Polynomial([[2, 'x']]), 5, Polynomial([[2**5, 'x', 'x', 'x', 'x', 'x']])),
    (RationalPolynomial([[2, 'x']]), 5, RationalPolynomial([[2**5, 'x', 'x', 'x', 'x', 'x']])),
])
def test_pow(a, n, res):
    assert a ** n == res



"""Top-level package for Kingdon."""

__author__ = """Martin Roelfs"""
__email__ = 'martinroelfs@yahoo.com'
__version__ = '1.0.3'

from sympy import symbols

from kingdon.algebra import Algebra
from kingdon.multivector import MultiVector
from kingdon.matrixreps import expr_as_matrix



import pytest

from kingdon.operator_dict import OperatorDict, UnaryOperatorDict
from kingdon.codegen import codegen_gp, codegen_inv
from kingdon import Algebra


def test_operator_dict():
    alg = Algebra(2)
    x = alg.multivector(name='x')
    y = alg.multivector(name='y')

    gp = OperatorDict('gp', codegen=codegen_gp, algebra=alg)
    assert len(gp) == 0
    with pytest.raises(TypeError):
        gp[(x.keys(), y.keys())] = 2
    xy = gp(x, y)
    assert len(gp) == 1  # size of gp has grown by one
    assert (x.keys(), y.keys()) in gp

    inv = UnaryOperatorDict('inv', codegen=codegen_inv, algebra=alg)
    assert len(inv) == 0
    xinv = inv(x)
    assert len(inv) == 1
    assert x.keys() in inv



from kingdon import Algebra
from kingdon.matrixreps import expr_as_matrix
from sympy import Matrix, pprint


def test_expr_as_matrix():
    alg = Algebra(3, 0)
    x = alg.vector(name='x')
    B = alg.bivector(name='B')

    # Test for the matrix rep of the commutator. (Grade preserving)
    A, y = expr_as_matrix(alg.cp, B, x)
    B12, B13, B23 = B.values()
    assert A == Matrix([[0, B12, B13], [-B12, 0, B23], [-B13, -B23, 0]])
    assert y == B.cp(x)
    assert [alg.bin2canon[k] for k in y.keys()] == ['e1', 'e2', 'e3']

    # Test for the matrix rep of the commutator. (Grade preserving, only some output)
    X = alg.vector(e1=1, e3=1)
    A, y = expr_as_matrix(alg.cp, B, x, res_like=X)
    B12, B13, B23 = B.values()
    assert A == Matrix([[0, B12, B13], [-B13, -B23, 0]])
    assert [alg.bin2canon[k] for k in y.keys()] == ['e1', 'e3']

    # Test for the matrix rep of the anti-commutator.
    A, y = expr_as_matrix(alg.acp, B, x)
    assert A == Matrix([[B23, -B13, B12]])
    assert [alg.bin2canon[k] for k in y.keys()] == ['e123']

    # Test for the matrix rep of conjugation of the e3 plane.
    x = alg.vector(name='x', keys=('e3',))
    A, y = expr_as_matrix(alg.sw, B, x)
    assert A == Matrix([[2*B12*B23], [-2*B12*B13], [B12**2 - B13**2 - B23**2], [0]])
    assert [alg.bin2canon[k] for k in y.keys()] == ['e1', 'e2', 'e3', 'e123']



from __future__ import annotations

import string
from itertools import product, combinations, groupby
from collections import namedtuple, defaultdict
from typing import NamedTuple, Callable, Tuple, Dict
from functools import reduce, cached_property
import linecache
import warnings
import operator
from dataclasses import dataclass
import inspect
import builtins
import keyword

from sympy import simplify, sympify, Add, Mul, Symbol, expand
from sympy.utilities.iterables import iterable, flatten
from sympy.printing.lambdarepr import LambdaPrinter


@dataclass
class AdditionChains:
    limit: int

    @cached_property
    def minimal_chains(self) -> Dict[int, Tuple[int, ...]]:
        chains = {1: (1,)}
        while any(i not in chains for i in range(1, self.limit + 1)):
            for chain in chains.copy().values():
                right_summand = chain[-1]
                for left_summand in chain:
                    value = left_summand + right_summand
                    if value <= self.limit and value not in chains:
                        chains[value] = (*chain, value)
        return chains

    def __getitem__(self, n: int) -> Tuple[int, ...]:
        return self.minimal_chains[n]

    def __contains__(self, item):
        return self[item]

def power_supply(x: "MultiVector", exponents: Tuple[int, ...], operation: Callable[["MultiVector", "MultiVector"], "MultiVector"] = operator.mul):
    """
    Generates powers of a given multivector using the least amount of multiplications.
    For example, to raise a multivector :math:`x` to the power :math:`a = 15`, only 5
    multiplications are needed since :math:`x^{2} = x * x`, :math:`x^{3} = x * x^2`,
    :math:`x^{5} = x^2 * x^3`, :math:`x^{10} = x^5 * x^5`, :math:`x^{15} = x^5 * x^{10}`.
    The :class:`power_supply` uses :class:`AdditionChains` to determine these shortest
    chains.

    When called with only a single integer, e.g. :code:`power_supply(x, 15)`, iterating
    over it yields the above sequence in order; ending with :math:`x^{15}`.

    When called with a sequence of integers, the generator instead returns only the requested terms.


    :param x: The MultiVector to be raised to a power.
    :param exponents: When an :code:`int`, this generates the shortest possible way to
        get to :math:`x^a`, where :math:`x`
    """
    if isinstance(exponents, int):
        target = exponents
        addition_chains = AdditionChains(target)
        exponents = addition_chains[target]
    else:
        addition_chains = AdditionChains(max(exponents))

    powers = {1: x}
    for step in exponents:
        if step not in powers:
            chain = addition_chains[step]
            powers[step] = operation(powers[chain[-2]], powers[step - chain[-2]])

        yield powers[step]


class TermTuple(NamedTuple):
    """
    TermTuple represents a single monomial in a product of multivectors.

    :param key_out: is the basis blade to which this monomial belongs.
    :param keys_in: are the input basis blades in this monomial.
    :param sign: Sign of the monomial.
    :param values_in: Input values. Typically, tuple of :class:`sympy.core.symbol.Symbol`.
    :param termstr: The string representation of this monomial, e.g. '-x*y'.
    """
    key_out: int
    keys_in: Tuple[int]
    sign: int
    values_in: Tuple["sympy.core.symbol.Symbol"]
    termstr: str


class CodegenOutput(NamedTuple):
    """
    Output of a codegen function.

    :param keys_out: tuple with the output blades in binary rep.
    :param func: callable that takes (several) sequence(s) of values
        returns a tuple of :code:`len(keys_out)`.
    """
    keys_out: Tuple[int]
    func: Callable


def term_tuple(items, sign_func, keyout_func=operator.xor):
    """
    Create a single term in a multivector product between the basis blades present in `items`.
    """
    keys_in, values_in = zip(*items)
    sign = sign_func(keys_in)
    if not sign:
        return TermTuple(key_out=0, keys_in=keys_in, sign=sign, values_in=values_in, termstr='')
    key_out = keyout_func(*keys_in)
    return TermTuple(key_out, keys_in, sign, values_in,
                     f'{"+" if sign > 0 else "-"}{"*".join(str(v) for v in values_in)}')


def codegen_product(x, y, filter_func=None, sign_func=None, keyout_func=operator.xor):
    """
    Helper function for the codegen of all product-type functions.

    :param x: Fully symbolic :class:`~kingdon.multivector.MultiVector`.
    :param y: Fully symbolic :class:`~kingdon.multivector.MultiVector`.
    :param filter_func: A condition which should be true in the preprocessing of terms.
        Input is a TermTuple.
    :param sign_func: function to compute sign between terms. E.g. algebra.signs[ei, ej]
        for metric dependent products. Input: 2-tuple of blade indices, e.g. (ei, ej).
    :param keyout_func:
    """
    algebra = x.algebra
    if sign_func is None:
        sign_func = lambda pair: algebra.signs[pair]

    # If sign == 0, then the term should be disregarded since it is zero
    terms = filter(lambda tt: tt.sign, (term_tuple(items, sign_func, keyout_func=keyout_func)
                                        for items in product(x.items(), y.items())))
    if filter_func is not None:
        terms = filter(filter_func, terms)

    res = defaultdict(str)
    for term in terms:
        if term.key_out in res:
            res[term.key_out] += term.termstr
        else:
            res[term.key_out] = term.termstr[1:] if term.termstr[0] == '+' else term.termstr
    return dict(sorted(res.items()))


def codegen_gp(x, y):
    """
    Generate the geometric product between :code:`x` and :code:`y`.

    :param x: Fully symbolic :class:`~kingdon.multivector.MultiVector`.
    :param y: Fully symbolic :class:`~kingdon.multivector.MultiVector`.
    :return: tuple with integers indicating the basis blades present in the
        product in binary convention, and a lambda function that perform the product.
    """
    return codegen_product(x, y)


def codegen_sw(x, y):
    """
    Generate the conjugation of :code:`y` by :code:`x`: :math:`x y \widetilde{x}`.

    :return: tuple of keys in binary representation and a lambda function.
    """
    return x * y * ~x


def codegen_cp(x, y):
    """
    Generate the commutator product of :code:`x` and :code:`y`: :code:`x.cp(y) = 0.5*(x*y-y*x)`.

    :return: tuple of keys in binary representation and a lambda function.
    """
    algebra = x.algebra
    filter_func = lambda tt: (algebra.signs[tt.keys_in] - algebra.signs[tt.keys_in[::-1]])
    return codegen_product(x, y, filter_func=filter_func)


def codegen_acp(x, y):
    """
    Generate the anti-commutator product of :code:`x` and :code:`y`: :code:`x.acp(y) = 0.5*(x*y+y*x)`.

    :return: tuple of keys in binary representation and a lambda function.
    """
    algebra = x.algebra
    filter_func = lambda tt: (algebra.signs[tt.keys_in] + algebra.signs[tt.keys_in[::-1]])
    return codegen_product(x, y, filter_func=filter_func)


def codegen_ip(x, y, diff_func=abs):
    """
    Generate the inner product of :code:`x` and :code:`y`.

    :param diff_func: How to treat the difference between the binary reps of the basis blades.
        if :code:`abs`, compute the symmetric inner product. When :code:`lambda x: -x` this
        function generates left-contraction, and when :code:`lambda x: x`, right-contraction.
    :return: tuple of keys in binary representation and a lambda function.
    """
    algebra = x.algebra
    filter_func = lambda tt: tt.key_out == diff_func(tt.keys_in[0] - tt.keys_in[1])
    return codegen_product(x, y, filter_func=filter_func)


def codegen_lc(x, y):
    """
    Generate the left-contraction of :code:`x` and :code:`y`.

    :return: tuple of keys in binary representation and a lambda function.
    """
    return codegen_ip(x, y, diff_func=lambda x: -x)


def codegen_rc(x, y):
    """
    Generate the right-contraction of :code:`x` and :code:`y`.

    :return: tuple of keys in binary representation and a lambda function.
    """
    return codegen_ip(x, y, diff_func=lambda x: x)


def codegen_sp(x, y):
    """
    Generate the scalar product of :code:`x` and :code:`y`.

    :return: tuple of keys in binary representation and a lambda function.
    """
    return codegen_ip(x, y, diff_func=lambda x: 0)


def codegen_proj(x, y):
    """
    Generate the projection of :code:`x` onto :code:`y`: :math:`(x \cdot y) \widetilde{y}`.

    :return: tuple of keys in binary representation and a lambda function.
    """
    return (x | y) * ~y


def codegen_op(x, y):
    """
    Generate the outer product of :code:`x` and :code:`y`: :code:`x.op(y) = x ^ y`.

    :x: MultiVector
    :y: MultiVector
    :return: dictionary with integer keys indicating the corresponding basis blade in binary convention,
        and values which are a 3-tuple of indices in `x`, indices in `y`, and a lambda function.
    """
    algebra = x.algebra
    filter_func = lambda tt: tt.key_out == sum(tt.keys_in)
    sign_func = lambda pair: (-1)**algebra.swaps[pair]
    return codegen_product(x, y, filter_func=filter_func, sign_func=sign_func)


def codegen_rp(x, y):
    """
    Generate the regressive product of :code:`x` and :code:`y`:,
    :math:`x \\vee y`.

    :param x:
    :param y:
    :return: tuple of keys in binary representation and a lambda function.
    """
    algebra = x.algebra
    keyout_func = lambda tot, key_in: len(algebra) - 1 - (key_in ^ tot)
    filter_func = lambda tt: len(algebra) - 1 == sum(tt.keys_in) - tt.key_out
    # Sign is composed of dualization of each blade, exterior product, and undual.
    sign_func = lambda pair: (
        algebra.signs[pair[0], len(algebra) - 1 - pair[0]] *
        algebra.signs[pair[1], len(algebra) - 1 - pair[1]] *
        (-1)**algebra.swaps[len(algebra) - 1 - pair[0], len(algebra) - 1 - pair[1]] *
        algebra.signs[len(algebra) - 1 - (pair[0] ^ pair[1]), pair[0] ^ pair[1]]
    )

    return codegen_product(
        x, y,
        filter_func=filter_func,
        keyout_func=keyout_func,
        sign_func=sign_func,
    )


Fraction = namedtuple('Fraction', ['numer', 'denom'])
Fraction.__doc__ = """
Tuple representing a fraction.
"""


class LambdifyInput(NamedTuple):
    """ Strike package for the Lambdify function. """
    funcname: str
    args: dict
    expr_dict: dict
    dependencies: list


def codegen_inv(y, x=None, symbolic=False):
    alg = y.algebra
    if alg.d < 6:
        num, denom = codegen_hitzer_inv(y, symbolic=True)
    else:
        num, denom = codegen_shirokov_inv(y, symbolic=True)
    num = num if x is None else x * num

    if symbolic:
        return Fraction(num, denom)

    d = alg.scalar(name='d', symbolcls=alg.codegen_symbolcls)
    denom_inv = alg.scalar([1 / denom])
    yinv = num * d.e  #TODO: this multiply is too gready, would be better if it didnt distribute, & reinstate CSE

    # Prepare all the input for lambdify
    args = {'y': y.values()}
    expr_dict = dict(yinv.items())
    dependencies = list(zip(d.values(), denom_inv.values()))
    return LambdifyInput(
        funcname=f'codegen_inv_{y.type_number}',
        expr_dict=expr_dict,
        args=args,
        dependencies=dependencies,
    )

def codegen_hitzer_inv(x, symbolic=False):
    """
    Generate code for the inverse of :code:`x` using the Hitzer inverse,
    which works up to 5D algebras.
    """
    alg = x.algebra
    d = alg.d
    if d == 0:
        num = alg.blades.e
    elif d == 1:
        num = x.involute()
    elif d == 2:
        num = x.conjugate()
    elif d == 3:
        xconj = x.conjugate()
        num = xconj * ~(x * xconj)
    elif d == 4:
        xconj = x.conjugate()
        x_xconj = x * xconj
        num = xconj * (x_xconj - 2 * x_xconj.grade(3, 4))
    elif d == 5:
        xconj = x.conjugate()
        x_xconj = x * xconj
        combo = xconj * ~x_xconj
        x_combo = x * combo
        num = combo * (x_combo - 2 * x_combo.grade(1, 4))
    else:
        raise NotImplementedError(f"Closed form inverses are not known in {d=} dimensions.")
    denom = (x * num).e

    if symbolic:
        return Fraction(num, denom)
    return alg.multivector({k: v / denom for k, v in num.items()})

def codegen_shirokov_inv(x, symbolic=False):
    """
    Generate code for the inverse of :code:`x` using the Shirokov inverse,
    which is works in any algebra, but it can be expensive to compute.
    """
    alg = x.algebra
    n = 2 ** ((alg.d + 1) // 2)
    supply = power_supply(x, tuple(range(1, n + 1)))  # Generate powers of x efficiently.
    powers = []
    cs = []
    xs = []
    for i in range(1, n + 1):
        powers.append(next(supply))
        xi = powers[i - 1]
        for j in range(i - 1):
            power_idx = i - j - 2
            xi_diff = powers[power_idx] * cs[j]
            xi = xi - xi_diff
        if xi.grades == (0,):
            break
        xs.append(xi)
        cs.append(s if (s := xi.e) == 0 else n * s / i)

    if i == 1:
        adj = alg.blades.e
    else:
        adj = xs[-1] - cs[-1]

    if symbolic:
        return Fraction(adj, xi.e)
    return alg.multivector({k: v / xi.e for k, v in adj.items()})


def codegen_div(x, y):
    """
    Generate code for :math:`x y^{-1}`.
    """
    alg = x.algebra
    num, denom = codegen_inv(y, x, symbolic=True)
    if not denom:
        raise ZeroDivisionError
    d = alg.scalar(name='d', symbolcls=alg.codegen_symbolcls)
    denom_inv = alg.scalar([1 / denom])
    res = num * d.e

    # Prepare all the input for lambdify
    args = {'x': x.values(), 'y': y.values()}
    expr_dict = dict(res.items())
    dependencies = list(zip(d.values(), denom_inv.values()))
    return LambdifyInput(
        funcname=f'div_{x.type_number}_x_{y.type_number}',
        expr_dict=expr_dict,
        args=args,
        dependencies=dependencies,
    )


def codegen_normsq(x):
    return x * ~x


def codegen_outerexp(x, asterms=False):
    alg = x.algebra
    if len(x.grades) != 1:
        warnings.warn('Outer exponential might not converge for mixed-grade multivectors.', RuntimeWarning)
    k = alg.d

    Ws = [alg.scalar([1]), x]
    j = 2
    while j <= k:
        Wj = Ws[-1] ^ x
        # Dividing like this avoids floating point numbers, which is excellent.
        Wj._values = tuple(v / j for v in Wj._values)
        if Wj:
            Ws.append(Wj)
            j += 1
        else:
            break

    if asterms:
        return Ws
    return reduce(operator.add, Ws)

def codegen_outersin(x):
    odd_Ws = codegen_outerexp(x, asterms=True)[1::2]
    outersin = reduce(operator.add, odd_Ws)
    return outersin


def codegen_outercos(x):
    even_Ws = codegen_outerexp(x, asterms=True)[0::2]
    outercos = reduce(operator.add, even_Ws)
    return outercos


def codegen_outertan(x):
    Ws = codegen_outerexp(x, asterms=True)
    even_Ws, odd_Ws = Ws[0::2], Ws[1::2]
    outercos = reduce(operator.add, even_Ws)
    outersin = reduce(operator.add, odd_Ws)
    outertan = outersin / outercos
    return outertan


def codegen_add(x, y):
    vals = dict(x.items())
    for k, v in y.items():
        if k in vals:
            vals[k] = vals[k] + v
        else:
            vals[k] = v
    return vals


def codegen_sub(x, y):
    vals = dict(x.items())
    for k, v in y.items():
        if k in vals:
            vals[k] = vals[k] - v
        else:
            vals[k] = -v
    return vals

def codegen_neg(x):
    return {k: -v for k, v in x.items()}


def codegen_involutions(x, invert_grades=(2, 3)):
    """
    Codegen for the involutions of Clifford algebras:
    reverse, grade involute, and Clifford involution.

    :param invert_grades: The grades that flip sign under this involution mod 4, e.g. (2, 3) for reversion.
    """
    return {k: -v if bin(k).count('1') % 4 in invert_grades else v
            for k, v in x.items()}


def codegen_reverse(x):
    return codegen_involutions(x, invert_grades=(2, 3))


def codegen_involute(x):
    return codegen_involutions(x, invert_grades=(1, 3))


def codegen_conjugate(x):
    return codegen_involutions(x, invert_grades=(1, 2))


def codegen_sqrt(x):
    """
    Take the square root using the study number approach as described in
    https://doi.org/10.1002/mma.8639
    """
    alg = x.algebra
    if x.grades == (0,):
        return {0: f'({str(x.e)}**0.5)'}
    a, bI = x.grade(0), x - x.grade(0)
    has_solution = len(x.grades) <= 2 and 0 in x.grades
    if not has_solution:
        warnings.warn("Cannot verify that we really are taking the sqrt of a Study number.", RuntimeWarning)

    bI_sq = bI * bI
    if not bI_sq:
        cp = f'({str(a.e)}**0.5)'
    else:
        normS = (a * a - bI * bI).e
        cp = f'(0.5 * ({str(a.e)} + {str(normS)}**0.5)) ** 0.5'
    c = alg.scalar(name='c')
    c2_inv = alg.scalar(name='c2_inv')
    dI = bI * c2_inv
    res = c + dI

    # Prepare all the input for lambdify
    args = {'x': x.values()}
    expr_dict = dict(res.items())
    dependencies = [*zip(c.values(), [cp]), *zip(c2_inv.values(), [f'0.5 / {cp}'])]
    return LambdifyInput(
        funcname=f'sqrt_{x.type_number}',
        expr_dict=expr_dict,
        args=args,
        dependencies=dependencies,
    )


def codegen_polarity(x, undual=False):
    if undual:
        return x * x.algebra.pss
    sign = x.algebra.signs[-1, -1]
    if sign == 1:
        return x * x.algebra.pss
    if sign == -1:
        return x * (-x.algebra.pss)
    if sign == 0:
        raise ZeroDivisionError


def codegen_unpolarity(x):
    return codegen_polarity(x, undual=True)


def codegen_hodge(x, undual=False):
    if undual:
        return x.algebra.multivector(
            {len(x.algebra) - 1 - eI: x.algebra.signs[len(x.algebra) - 1 - eI, eI] * val
             for eI, val in x.items()}
        )
    return x.algebra.multivector(
        {len(x.algebra) - 1 - eI: x.algebra.signs[eI, len(x.algebra) - 1 - eI] * val
         for eI, val in x.items()}
    )


def codegen_unhodge(x):
    return codegen_hodge(x, undual=True)


def _lambdify_mv(mv):
    func = lambdify(
        args={'x': sorted(mv.free_symbols, key=lambda x: x.name)},
        exprs=list(mv.values()),
        funcname=f'custom_{mv.type_number}',
        cse=mv.algebra.cse
    )
    return CodegenOutput(tuple(mv.keys()), func)


def do_codegen(codegen, *mvs) -> CodegenOutput:
    """
    :param codegen: callable that performs codegen for the given :code:`mvs`. This can be any callable
        that returns either a :class:`~kingdon.multivector.MultiVector`, a dictionary, or an instance of :class:`CodegenOutput`.
    :param mvs: Any remaining positional arguments are taken to be symbolic :class:`~kingdon.multivector.MultiVector`'s.
    :return: Instance of :class:`CodegenOutput`.
    """
    algebra = mvs[0].algebra
    namespace = algebra.numspace

    res = codegen(*mvs)

    if isinstance(res, CodegenOutput):
        return res

    if isinstance(res, LambdifyInput):
        funcname = res.funcname
        args = res.args
        dependencies = res.dependencies
        res = res.expr_dict
    else:
        funcname = f'{codegen.__name__}_' + '_x_'.join(f"{mv.type_number}" for mv in mvs)
        args = {arg_name: arg.values() for arg_name, arg in zip(string.ascii_uppercase, mvs)}
        dependencies = None

    # Sort the keys in canonical order
    res = {bin: res[bin] if isinstance(res, dict) else getattr(res, canon)
           for canon, bin in algebra.canon2bin.items() if bin in res.keys()}

    if not algebra.cse and any(isinstance(v, str) for v in res.values()):
        return func_builder(res, *mvs, funcname=funcname)


    keys, exprs = tuple(res.keys()), list(res.values())
    func = lambdify(args, exprs, funcname=funcname, cse=algebra.cse, dependencies=dependencies)
    return CodegenOutput(
        keys, func
    )

def do_compile(codegen, *tapes):
    algebra = tapes[0].algebra
    namespace = algebra.numspace

    res = codegen(*tapes)
    funcname = f'{codegen.__name__}_' + '_x_'.join(f"{tape.type_number}" for tape in tapes)
    funcstr = f"def {funcname}({', '.join(t.expr for t in tapes)}):"
    if not isinstance(res, str):
        funcstr += f"    return {res.expr}"
    else:
        funcstr += f"    return ({res},)"

    funclocals = {}
    filename = f'<{funcname}>'
    c = compile(funcstr, filename, 'exec')
    exec(c, namespace, funclocals)
    # mtime has to be None or else linecache.checkcache will remove it
    linecache.cache[filename] = (len(funcstr), None, funcstr.splitlines(True), filename) # type: ignore

    func = funclocals[funcname]
    return CodegenOutput(
        res.keys() if not isinstance(res, str) else (0,), func
    )


def func_builder(res_vals: defaultdict, *mvs, funcname: str) -> CodegenOutput:
    """
    Build a Python function for the product between given multivectors.

    :param res_vals: Dict to be converted into a function. The keys correspond to the basis blades in binary,
        while the values are strings to be converted into source code.
    :param mvs: all the multivectors that the resulting function is a product of.
    :param funcname: Name of the function. Be aware: if a function by that name already existed, it will be overwritten.
    :return: tuple of output keys of the callable, and the callable.
    """
    args = string.ascii_uppercase[:len(mvs)]
    header = f'def {funcname}({", ".join(args)}):'
    if res_vals:
        body = ''
        for mv, arg in zip(mvs, args):
            body += f'    [{", ".join(str(v) for v in mv.values())}] = {arg}\n'
        return_val = f'    return [{", ".join(res_vals.values())},]'
    else:
        body = ''
        return_val = f'    return list()'
    func_source = f'{header}\n{body}\n{return_val}'

    # Dynamically build a function
    func_locals = {}
    c = compile(func_source, funcname, 'exec')
    exec(c, {}, func_locals)

    # Add the generated code to linecache such that it is inspect-safe.
    linecache.cache[funcname] = (len(func_source), None, func_source.splitlines(True), funcname)
    func = func_locals[funcname]
    return CodegenOutput(tuple(res_vals.keys()), func)


def lambdify(args: dict, exprs: list, funcname: str, dependencies: tuple = None, printer=LambdaPrinter, dummify=False, cse=False):
    """
    Function that turns symbolic expressions into Python functions. Heavily inspired by
    :mod:`sympy`'s function by the same name, but adapted for the needs of :code:`kingdon`.

    Particularly, this version gives us more control over the names of the function and its
    arguments, and is more performant, particularly when the given expressions are strings.

    Example usage:

    .. code-block ::

        alg = Algebra(2)
        a = alg.multivector(name='a')
        b = alg.multivector(name='b')
        args = {'A': a.values(), 'B': b.values()}
        exprs = tuple(codegen_cp(a, b).values())
        func = lambdify(args, exprs, funcname='cp', cse=False)

    This will produce the following code:

    .. code-block ::

        def cp(A, B):
            [a, a1, a2, a12] = A
            [b, b1, b2, b12] = B
            return (+a1*b2-a2*b1,)

    It is recommended not to call this function directly, but rather to use
    :func:`do_codegen` which provides a clean API around this function.

    :param args: dictionary of type dict[str | Symbol, tuple[Symbol]].
    :param exprs: tuple[Expr]
    :param funcname: string to be used as the bases for the name of the function.
    :param dependencies: These are extra expressions that can be provided such that quantities can be precomputed.
        For example, in the inverse of a multivector, this is used to compute the scalar denominator only once,
        after which all values in expr are multiplied by it. When :code:`cse = True`, these dependencies are also
        included in the CSE process.
    :param cse: If :code:`True` (default), CSE is applied to the expressions and dependencies.
        This typically greatly improves performance and reduces numba's initialization time.
    :return: Function that represents that can be used to calculate the values of exprs.
    """
    if printer is LambdaPrinter:
        printer = LambdaPrinter(
            {'fully_qualified_modules': False, 'inline': True,
             'allow_unknown_functions': True,
             'user_functions': {}}
        )

    tosympy = lambda x: x.tosympy() if hasattr(x, 'tosympy') else x
    args = {name: [tosympy(v) for v in values]
            for name, values in args.items()}
    exprs = [tosympy(expr) for expr in exprs]
    if dependencies is not None:
        dependencies = [(tosympy(y), tosympy(x)) for y, x in dependencies]
    names = tuple(arg if isinstance(arg, str) else arg.name for arg in args.keys())
    iterable_args = tuple(args.values())

    funcprinter = KingdonPrinter(printer, dummify)

    # TODO: Extend CSE to include the dependencies.
    lhsides, rhsides = zip(*dependencies) if dependencies else ([], [])
    if cse and not any(isinstance(expr, str) for expr in exprs):
        if not callable(cse):
            from sympy.simplify.cse_main import cse
        if dependencies:
            all_exprs = [*exprs, *rhsides]
            cses, _all_exprs = cse(all_exprs, list=False, order='none', ignore=lhsides)
            _exprs, _rhsides = _all_exprs[:-len(rhsides)], _all_exprs[len(exprs):]
            cses.extend(list(zip(flatten(lhsides), flatten(_rhsides))))
        else:
            cses, _exprs = cse(exprs, list=False)
    else:
        cses, _exprs = list(zip(flatten(lhsides), flatten(rhsides))), exprs

    if not any(_exprs):
        _exprs = list('0' for expr in _exprs)
    funcstr = funcprinter.doprint(funcname, iterable_args, names, _exprs, cses=cses)

    # Provide lambda expression with builtins, and compatible implementation of range
    namespace = {'builtins': builtins, 'range': range}

    funclocals = {}
    filename = f'<{funcname}>'
    c = compile(funcstr, filename, 'exec')
    exec(c, namespace, funclocals)
    # mtime has to be None or else linecache.checkcache will remove it
    linecache.cache[filename] = (len(funcstr), None, funcstr.splitlines(True), filename) # type: ignore

    func = funclocals[funcname]
    return func


class KingdonPrinter:
    def __init__(self, printer=None, dummify=False):
        self._dummify = dummify

        #XXX: This has to be done here because of circular imports
        from sympy.printing.lambdarepr import LambdaPrinter

        if printer is None:
            printer = LambdaPrinter()

        if inspect.isfunction(printer):
            self._exprrepr = printer
        else:
            if inspect.isclass(printer):
                printer = printer()

            self._exprrepr = printer.doprint

        # Used to print the generated function arguments in a standard way
        self._argrepr = LambdaPrinter().doprint

    def doprint(self, funcname, args, names, expr, *, cses=()):
        """
        Returns the function definition code as a string.
        """
        funcbody = []

        if not iterable(args):
            args = [args]

        if cses:
            subvars, subexprs = zip(*cses)
            exprs = [expr] + list(subexprs)
            argstrs, exprs = self._preprocess(args, exprs)
            expr, subexprs = exprs[0], exprs[1:]
            cses = zip(subvars, subexprs)
        else:
            argstrs, expr = self._preprocess(args, expr)

        # Generate argument unpacking and final argument list
        funcargs = []
        unpackings = []

        for name, argstr in zip(names, argstrs):
            if iterable(argstr):
                funcargs.append(name)
                unpackings.extend(self._print_unpacking(argstr, funcargs[-1]))
            else:
                funcargs.append(argstr)

        funcsig = 'def {}({}):'.format(funcname, ', '.join(funcargs))

        # Wrap input arguments before unpacking
        funcbody.extend(self._print_funcargwrapping(funcargs))

        funcbody.extend(unpackings)

        for s, e in cses:
            if e is None:
                funcbody.append('del {}'.format(s))
            else:
                funcbody.append('{} = {}'.format(s, self._exprrepr(e)))

        str_expr = _recursive_to_string(self._exprrepr, expr)

        if '\n' in str_expr:
            str_expr = '({})'.format(str_expr)
        funcbody.append('return {}'.format(str_expr))

        funclines = [funcsig]
        funclines.extend(['    ' + line for line in funcbody])

        return '\n'.join(funclines) + '\n'

    @classmethod
    def _is_safe_ident(cls, ident):
        return isinstance(ident, str) and ident.isidentifier() \
                and not keyword.iskeyword(ident)

    def _preprocess(self, args, expr):
        """Preprocess args, expr to replace arguments that do not map
        to valid Python identifiers.

        Returns string form of args, and updated expr.
        """
        argstrs = [None]*len(args)
        for i, arg in enumerate(args):
            if iterable(arg):
                s, expr = self._preprocess(arg, expr)
            elif hasattr(arg, 'name'):
                s = arg.name
            elif hasattr(arg, 'is_symbol') and arg.is_symbol:
                s = self._argrepr(arg)
            else:
                s = str(arg)
            argstrs[i] = s
        return argstrs, expr

    def _print_funcargwrapping(self, args):
        """Generate argument wrapping code.

        args is the argument list of the generated function (strings).

        Return value is a list of lines of code that will be inserted  at
        the beginning of the function definition.
        """
        return []

    def _print_unpacking(self, unpackto, arg):
        """Generate argument unpacking code.

        arg is the function argument to be unpacked (a string), and
        unpackto is a list or nested lists of the variable names (strings) to
        unpack to.
        """
        def unpack_lhs(lvalues):
            return '[{}]'.format(', '.join(
                unpack_lhs(val) if iterable(val) else val for val in lvalues))

        return ['{} = {}'.format(unpack_lhs(unpackto), arg)]

def _recursive_to_string(doprint, arg):
    if isinstance(arg, str):
        return arg
    elif not arg:
        return str(arg)  # Empty list or tuple
    elif iterable(arg):
        if isinstance(arg, list):
            left, right = "[", "]"
        elif isinstance(arg, tuple):
            left, right = "(", ",)"
        else:
            raise NotImplementedError("unhandled type: %s, %s" % (type(arg), arg))
        return ''.join((left, ', '.join(_recursive_to_string(doprint, e) for e in arg), right))
    else:
        return doprint(arg)



from dataclasses import dataclass, field
from functools import cached_property, partial, partialmethod


@dataclass(init=False)
class TapeRecorder:
    #TODO: Create common baseclass for TapeRecorder and Multivector?
    algebra: "Algebra"
    expr: str
    _keys: tuple = field(default_factory=tuple)

    def __new__(cls, algebra, expr, keys):
        obj = object.__new__(cls)
        obj.algebra = algebra
        obj.expr = expr
        obj._keys = keys
        return obj

    def keys(self):
        return self._keys

    @cached_property
    def type_number(self) -> int:
        return int(''.join('1' if i in self.keys() else '0' for i in reversed(self.algebra.canon2bin.values())), 2)

    def __getattr__(self, basis_blade):
        bin_blade = self.algebra.canon2bin[basis_blade]
        try:
            idx = self.keys().index(bin_blade)
        except ValueError:
            return self.__class__(
                algebra=self.algebra,
                expr=f"(0,)",
                keys=(0,)
            )
        else:
            return self.__class__(
                algebra=self.algebra,
                expr=f"({self.expr}[{idx}],)",
                keys=(self.keys()[idx],)
            )

    def grade(self, *grades):
        if len(grades) == 1 and isinstance(grades[0], tuple):
            grades = grades[0]

        basis_blades = self.algebra.indices_for_grades[grades]
        indices_keys = [(idx, k) for idx, k in enumerate(self.keys()) if k in basis_blades]
        indices, keys = zip(*indices_keys) if indices_keys else (tuple(), tuple())
        expr = f"[{self.expr}[idx] for idx in {indices}]"
        return self.__class__(
            algebra=self.algebra,
            expr=expr,
            keys=keys,
        )


    def __str__(self):
        return self.expr

    def binary_operator(self, other, operator: str):
        if not isinstance(other, self.__class__):
            # Assume scalar
            keys_out, func = getattr(self.algebra, operator)[self.keys(), (0,)]
            expr = f'{func.__name__}({self.expr}, ({other},))'
        else:
            keys_out, func = getattr(self.algebra, operator)[self.keys(), other.keys()]
            expr = f'{func.__name__}({self.expr}, {other.expr})'
        return self.__class__(algebra=self.algebra, expr=expr, keys=keys_out)

    def unary_operator(self, operator: str):
        keys_out, func = getattr(self.algebra, operator)[self.keys()]
        expr = f'{func.__name__}({self.expr})'
        return self.__class__(algebra=self.algebra, expr=expr, keys=keys_out)

    # Binary operators
    gp = __mul__ = __rmul__ = partialmethod(binary_operator, operator='gp')
    sw = __rshift__ = partialmethod(binary_operator, operator='sw')
    cp = partialmethod(binary_operator, operator='cp')
    acp = partialmethod(binary_operator, operator='acp')
    ip = __or__ = partialmethod(binary_operator, operator='ip')
    sp = partialmethod(binary_operator, operator='sp')
    lc = partialmethod(binary_operator, operator='lc')
    rc = partialmethod(binary_operator, operator='rc')
    op = __xor__ = __rxor__ = partialmethod(binary_operator, operator='op')
    rp = __and__ = partialmethod(binary_operator, operator='rp')
    proj = __matmul__ = partialmethod(binary_operator, operator='proj')
    add = __add__ = __radd__ = partialmethod(binary_operator, operator='add')
    sub = __sub__ = partialmethod(binary_operator, operator='sub')
    def __rsub__(self, other): return other + (-self)
    __truediv__ = div = partialmethod(binary_operator, operator='div')

    def __pow__(self, power, modulo=None):
        if power == 0:
            return self.__class__(self.algebra, expr='(1,)', keys=(0,))

        res = self
        for i in range(1, power):
            res = res.gp(self)
        return res

    # Unary operators
    inv = partialmethod(unary_operator, operator='inv')
    neg = __neg__ = partialmethod(unary_operator, operator='neg')
    reverse = __invert__ = partialmethod(unary_operator, operator='reverse')
    involute = partialmethod(unary_operator, operator='involute')
    conjugate = partialmethod(unary_operator, operator='conjugate')
    sqrt = partialmethod(unary_operator, operator='sqrt')
    polarity = partialmethod(unary_operator, operator='polarity')
    unpolarity = partialmethod(unary_operator, operator='unpolarity')
    hodge = partialmethod(unary_operator, operator='hodge')
    unhodge = partialmethod(unary_operator, operator='unhodge')
    normsq = partialmethod(unary_operator, operator='normsq')
    outerexp = partialmethod(unary_operator, operator='outerexp')
    outersin = partialmethod(unary_operator, operator='outersin')
    outercos = partialmethod(unary_operator, operator='outercos')
    outertan = partialmethod(unary_operator, operator='outertan')

    def dual(self, kind='auto'):
        if kind == 'polarity' or kind == 'auto' and self.algebra.r == 0:
            return self.polarity()
        elif kind == 'hodge' or kind == 'auto' and self.algebra.r == 1:
            return self.hodge()
        elif kind == 'auto':
            raise Exception('Cannot select a suitable dual in auto mode for this algebra.')
        else:
            raise ValueError(f'No dual found for kind={kind}.')

    def undual(self, kind='auto'):
        if kind == 'polarity' or kind == 'auto' and self.algebra.r == 0:
            return self.unpolarity()
        elif kind == 'hodge' or kind == 'auto' and self.algebra.r == 1:
            return self.unhodge()
        elif kind == 'auto':
            raise Exception('Cannot select a suitable undual in auto mode for this algebra.')
        else:
            raise ValueError(f'No undual found for kind={kind}.')



import operator
from collections.abc import Mapping
from dataclasses import dataclass, field
from functools import reduce, cached_property
from typing import Generator
from itertools import product

from sympy import Expr, Symbol, sympify

from kingdon.codegen import _lambdify_mv
from kingdon.polynomial import RationalPolynomial


@dataclass(init=False)
class MultiVector:
    algebra: "Algebra"
    _values: list = field(default_factory=list)
    _keys: tuple = field(default_factory=tuple)

    def __new__(cls, algebra: "Algebra", values=None, keys=None, *, name=None, grades=None, symbolcls=Symbol, **items):
        """
        :param algebra: Instance of :class:`~kingdon.algebra.Algebra`.
        :param keys: Keys corresponding to the basis blades in either binary rep or as strings, e.g. :code:'"e12"'.
        :param values: Values of the multivector. If keys are provided, then keys and values should
            satisfy :code:`len(keys) == len(values)`. If no keys nor grades are provided, :code:`len(values)`
            should equal :code:`len(algebra)`, i.e. a full multivector. If grades is provided,
            then :code:`len(values)` should be identical to the number of values in a multivector
            of that grade.
        :param name: Base string to be used as the name for symbolic values.
        :param grades: Optional, :class:`tuple` of grades in this multivector.
            If present, :code:`keys` is checked against these grades.
        :param symbolcls: Optional, class to be used for symbol creation. This is a :class:`sympy.Symbol` by default,
            but could be e.g. :class:`symfit.Variable` or :class:`symfit.Parameter` when the goal is to use this
            multivector in a fitting problem.
        :param items: keyword arguments can be used to initiate multivectors as well, e.g.
            :code:`MultiVector(alg, e12=1)`. Mutually exclusive with `values` and `keys`.
        """
        if items and keys is None and values is None:
            keys, values = zip(*((blade, items[blade]) for blade in algebra.canon2bin if blade in items))
            values = list(values)

        # Sanitize input
        if keys is not None and not all(isinstance(k, int) for k in keys):
            keys = tuple(k if k in algebra.bin2canon else algebra.canon2bin[k] for k in keys)
        if grades is None and name and keys is not None:
            grades = tuple(sorted({format(k, 'b').count('1') for k in keys}))
        values = values if values is not None else list()
        keys = keys if keys is not None else tuple()

        if grades is not None:
            if not all(0 <= grade <= algebra.d for grade in grades):
                raise ValueError(f'Each grade in `grades` needs to be a value between 0 and {algebra.d}.')
        else:
            if keys:
                grades = tuple(sorted({format(k, 'b').count('1') for k in keys}))
            else:
                grades = tuple(range(algebra.d + 1))

        if algebra.graded and keys and keys != algebra.indices_for_grades[grades]:
            raise ValueError(f"In graded mode, the keys should be equal to "
                             f"those expected for a multivector of {grades=}.")

        # Construct a new MV on the basis of the kind of input we received.
        if isinstance(values, Mapping):
            keys, values = zip(*values.items()) if values else (tuple(), list())
            values = list(values)
        elif len(values) == len(algebra.indices_for_grades[grades]) and not keys:
            keys = algebra.indices_for_grades[grades]
        elif name and not values:
            # values was not given, but we do have a name. So we are in symbolic mode.
            keys = algebra.indices_for_grades[grades] if not keys else keys
            values = list(symbolcls(f'{name}{algebra.bin2canon[k][1:]}') for k in keys)
        elif len(keys) != len(values):
            raise TypeError(f'Length of `keys` and `values` have to match.')

        if not all(isinstance(k, int) for k in keys):
            keys = tuple(key if key in algebra.bin2canon else algebra.canon2bin[key]
                         for key in keys)

        if any(isinstance(v, str) for v in values):
            values = list(val if not isinstance(val, str) else sympify(val)
                          for val in values)

        if not set(keys) <= set(algebra.indices_for_grades[grades]):
            raise ValueError(f"All keys should be of grades {grades}.")

        return cls.fromkeysvalues(algebra, keys, values)

    @classmethod
    def fromkeysvalues(cls, algebra, keys, values):
        """
        Initiate a multivector from a sequence of keys and a sequence of values.
        """
        obj = object.__new__(cls)
        obj.algebra = algebra
        obj._values = values
        obj._keys = keys
        return obj

    @classmethod
    def frommatrix(cls, algebra, matrix):
        """
        Initiate a multivector from a matrix. This matrix is assumed to be
        generated by :class:`~kingdon.multivector.MultiVector.asmatrix`, and
        thus we only read the first column of the input matrix.
        """
        obj = cls(algebra=algebra, values=matrix[..., 0])
        return obj

    def keys(self):
        return self._keys

    def values(self):
        return self._values

    def items(self):
        return zip(self._keys, self._values)

    def __len__(self):
        return len(self._values)

    @cached_property
    def type_number(self) -> int:
        return int(''.join('1' if i in self.keys() else '0' for i in reversed(self.algebra.canon2bin.values())), 2)


    def itermv(self, axis=None) -> Generator["MultiVector", None, None]:
        """
        Returns an iterator over the multivectors within this multivector, if it is a multidimensional multivector.
        For example, if you have a pointcloud of N points, itermv will iterate over these points one at a time.

        :param axis: Axis over which to iterate. Default is to iterate over all possible mv.
        """
        shape = self.shape[1:]
        if not shape:
            return self
        elif axis is None:
            return (
                self[indices]
                for indices in product(*(range(n) for n in shape))
            )
        else:
            raise NotImplementedError

    @property
    def shape(self):
        """ Return the shape of the .values() attribute of this multivector. """
        if hasattr(self._values, 'shape'):
            return self._values.shape
        elif hasattr(self._values[0], 'shape'):
            return len(self), *self._values[0].shape
        else:
            return len(self),

    @cached_property
    def grades(self):
        """ Tuple of the grades present in `self`. """
        return tuple(sorted({bin(ind).count('1') for ind in self.keys()}))

    def grade(self, *grades):
        """
        Returns a new  :class:`~kingdon.multivector.MultiVector` instance with
        only the selected `grades` from `self`.

        :param grades: tuple or ints, grades to select.
        """
        if len(grades) == 1 and isinstance(grades[0], tuple):
            grades = grades[0]

        vals = {k: getattr(self, self.algebra.bin2canon[k])
                for k in self.algebra.indices_for_grades[grades] if k in self.keys()}
        return self.fromkeysvalues(self.algebra, tuple(vals.keys()), list(vals.values()))

    @cached_property
    def issymbolic(self):
        """ True if this mv contains Symbols, False otherwise. """
        # Allowed symbol classes. codegen_symbolcls might refer to a constructor (method): get the class instead.
        symbol_classes = (Expr, self.algebra.codegen_symbolcls.__self__
                                if hasattr(self.algebra.codegen_symbolcls, '__self__')
                                else self.algebra.codegen_symbolcls)
        return any(isinstance(v, symbol_classes) for v in self.values())

    def neg(self):
        return self.algebra.neg(self)

    __neg__ = neg

    def __invert__(self):
        """ Reversion """
        return self.algebra.reverse(self)

    def reverse(self):
        """ Reversion """
        return self.algebra.reverse(self)

    def involute(self):
        """ Main grade involution. """
        return self.algebra.involute(self)

    def conjugate(self):
        """ Clifford conjugation: involution and reversion combined. """
        return self.algebra.conjugate(self)

    def sqrt(self):
        return self.algebra.sqrt(self)

    def normsq(self):
        return self.algebra.normsq(self)

    def norm(self):
        normsq = self.normsq()
        return normsq.sqrt()

    def normalized(self):
        """ Normalized version of this multivector. """
        return self / self.norm()

    def inv(self):
        """ Inverse of this multivector. """
        return self.algebra.inv(self)

    def add(self, other):
        return self.algebra.add(self, other)

    __radd__ = __add__ = add

    def sub(self, other):
        return self.algebra.sub(self, other)

    __sub__ = sub

    def __rsub__(self, other):
        return other + (-self)

    def div(self, other):
        return self.algebra.div(self, other)

    __truediv__ = div

    def __str__(self):
        if not len(self.values()):
            return '0'

        def print_value(val):
            if isinstance(val, Expr):
                if val.is_Symbol:
                    return f"{val}"
                else:
                    return f"({val})"
            elif isinstance(val, float):
                return f'{val:.3}'
            else:
                return f'{val}'

        canon_sorted_vals = {self.algebra._bin2canon_prettystr[key]: val for key, val in self.items()}
        str_repr = ' + '.join(
            [f'{print_value(val)} {blade}' if blade != '1' else f'{print_value(val)}'
             for blade, val in canon_sorted_vals.items() if (val.any() if hasattr(val, 'any') else val)]
        )
        return str_repr

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f'{self.__class__.__name__}(...)')
        else:
            p.text(str(self))

    def __format__(self, format_spec):
        if format_spec == 'keys_binary':
            iden = '_'.join(''.join('1' if i in self.keys() else '0' for i in bin_blades)
                            for bin_blades in self.algebra.indices_for_grade.values())
            return iden

    def __getitem__(self, item):
        if not isinstance(item, tuple):
            item = (item,)

        values = self.values()
        if isinstance(values, (tuple, list)):
            return_values = values.__class__(value[item] for value in values)
        else:
            return_values = values[(slice(None), *item)]
        return self.__class__.fromkeysvalues(self.algebra, keys=self.keys(), values=return_values)

    def __setitem__(self, indices, values):
        if isinstance(values, MultiVector):
            if self.keys() != values.keys():
                raise ValueError('setitem with a multivector is only possible for equivalent MVs.')
            values = values.values()

        if not isinstance(indices, tuple):
            indices = (indices,)

        if isinstance(self.values(), (tuple, list)):
            for self_values, other_value in zip(self.values(), values):
                self_values[indices] = other_value
        else:
            self.values()[(slice(None), *indices)] = values

    def __getattr__(self, basis_blade):
        # TODO: if this first check is not true, raise hell instead?
        if basis_blade not in self.algebra.canon2bin:
            return 0
        try:
            idx = self.keys().index(self.algebra.canon2bin[basis_blade])
        except ValueError:
            return 0
        return self._values[idx]

    def __contains__(self, item):
        item = item if isinstance(item, int) else self.algebra.canon2bin[item]
        return item in self._keys

    def __bool__(self):
        return bool(self.values())

    @cached_property
    def free_symbols(self):
        return reduce(operator.or_, (v.free_symbols for v in self.values() if hasattr(v, "free_symbols")))

    def map(self, func) -> "MultiVector":
        """ Returns a new multivector where `func` has been applied to all the values."""
        vals = [func(v) for v in self.values()]
        return self.fromkeysvalues(self.algebra, keys=self.keys(), values=vals)

    def filter(self, func=None) -> "MultiVector":
        """
        Returns a new multivector containing only those elements for which `func` was true-ish.
        If no function was provided, use the simp_func of the Algebra.
        """
        if func is None:
            func = self.algebra.simp_func
        keysvalues = tuple((k, v) for k, v in self.items() if func(v))
        if not keysvalues:
            return self.fromkeysvalues(self.algebra, keys=tuple(), values=list())
        keys, values = zip(*keysvalues)
        return self.fromkeysvalues(self.algebra, keys=keys, values=list(values))

    @cached_property
    def _callable(self):
        """ Return the callable function for this MV. """
        return _lambdify_mv(self)

    def __call__(self, *args, **kwargs):
        if args and kwargs:
            raise Exception('Please provide all input either as positional arguments or as keywords arguments, not both.')

        if not self.free_symbols:
            return self
        keys_out, func = self._callable
        if kwargs:
            args = [v for k, v in sorted(kwargs.items(), key=lambda x: x[0])]
        values = func(args)
        return self.fromkeysvalues(self.algebra, keys_out, values)

    def asmatrix(self):
        """ Returns a matrix representation of this multivector. """
        bin2index = {k: i for i, k in enumerate(self.algebra.canon2bin.values())}
        return sum(v * self.algebra.matrix_basis[bin2index[k]] for k, v in self.items())

    def asfullmv(self, canonical=True):
        """
        Returns a full version of the same multivector.

        :param canonical: If True (default) the values are in canonical order,
          even if the mutivector was already dense.
        """
        if canonical:
            keys = self.algebra.indices_for_grades[tuple(range(self.algebra.d + 1))]
        else:
            keys = tuple(range(len(self.algebra)))
        values = [getattr(self, self.algebra.bin2canon[k]) for k in keys]
        return self.fromkeysvalues(self.algebra, keys=keys, values=values)

    def gp(self, other):
        return self.algebra.gp(self, other)

    __mul__ = gp

    def __rmul__(self, other):
        return self.algebra.gp(other, self)

    def sw(self, other):
        """
        Apply :code:`x := self` to :code:`y := other` under conjugation:
        :code:`x.sw(y) = x*y*~x`.
        """
        return self.algebra.sw(self, other)

    __rshift__ = sw

    def __rrshift__(self, other):
        return self.algebra.sw(other, self)

    def proj(self, other):
        """
        Project :code:`x := self` onto :code:`y := other`: :code:`x @ y = (x | y) * ~y`.
        For correct behavior, :code:`x` and :code:`y` should be normalized (k-reflections).
        """
        return self.algebra.proj(self, other)

    __matmul__ = proj

    def __rmatmul__(self, other):
        return self.algebra.proj(other, self)

    def cp(self, other):
        """
        Calculate the commutator product of :code:`x := self` and :code:`y := other`:
        :code:`x.cp(y) = 0.5*(x*y-y*x)`.
        """
        return self.algebra.cp(self, other)

    def acp(self, other):
        """
        Calculate the anti-commutator product of :code:`x := self` and :code:`y := other`:
        :code:`x.cp(y) = 0.5*(x*y+y*x)`.
        """
        return self.algebra.acp(self, other)

    def ip(self, other):
        return self.algebra.ip(self, other)

    __or__ = ip

    def __ror__(self, other):
        return self.algebra.ip(other, self)

    def op(self, other):
        return self.algebra.op(self, other)

    __xor__ = __rxor__ = op

    def lc(self, other):
        return self.algebra.lc(self, other)

    def rc(self, other):
        return self.algebra.rc(self, other)

    def sp(self, other):
        """ Scalar product: :math:`\langle x \cdot y \rangle`. """
        return self.algebra.sp(self, other)

    def rp(self, other):
        return self.algebra.rp(self, other)

    __and__ = rp

    def __rand__(self, other):
        return self.algebra.rp(other, self)

    def __pow__(self, power, modulo=None):
        # TODO: this should also be taken care of via codegen, but for now this workaround is ok.
        if power == 0:
            return self.algebra.scalar((1,))

        res = self
        for i in range(1, power):
            res = res.gp(self)
        return res

    def outerexp(self):
        return self.algebra.outerexp(self)

    def outersin(self):
        return self.algebra.outersin(self)

    def outercos(self):
        return self.algebra.outercos(self)

    def outertan(self):
        return self.algebra.outertan(self)

    def polarity(self):
        return self.algebra.polarity(self)

    def unpolarity(self):
        return self.algebra.unpolarity(self)

    def hodge(self):
        return self.algebra.hodge(self)

    def unhodge(self):
        return self.algebra.unhodge(self)

    def dual(self, kind='auto'):
        """
        Compute the dual of `self`. There are three different kinds of duality in common usage.
        The first is polarity, which is simply multiplying by the inverse PSS. This is the only game in town for
        non-degenerate metrics (Algebra.r = 0). However, for degenerate spaces this no longer works, and we have
        two popular options: Poincaré and Hodge duality.

        By default, :code:`kingdon` will use polarity in non-degenerate spaces, and Hodge duality for spaces with
        `Algebra.r = 1`. For spaces with `r > 2`, little to no literature exists, and you are on your own.

        :param kind: if 'auto' (default), :code:`kingdon` will try to determine the best dual on the
            basis of the signature of the space. See explenation above.
            To ensure polarity, use :code:`kind='polarity'`, and to ensure Hodge duality,
            use :code:`kind='hodge'`.
        """
        if kind == 'polarity' or kind == 'auto' and self.algebra.r == 0:
            return self.polarity()
        elif kind == 'hodge' or kind == 'auto' and self.algebra.r == 1:
            return self.hodge()
        elif kind == 'auto':
            raise Exception('Cannot select a suitable dual in auto mode for this algebra.')
        else:
            raise ValueError(f'No dual found for kind={kind}.')

    def undual(self, kind='auto'):
        """
        Compute the undual of `self`. See :class:`~kingdon.multivector.MultiVector.dual` for more information.
        """
        if kind == 'polarity' or kind == 'auto' and self.algebra.r == 0:
            return self.unpolarity()
        elif kind == 'hodge' or kind == 'auto' and self.algebra.r == 1:
            return self.unhodge()
        elif kind == 'auto':
            raise Exception('Cannot select a suitable undual in auto mode for this algebra.')
        else:
            raise ValueError(f'No undual found for kind={kind}.')
